class BaseError(Exception):
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)
    def to_dict(self):
        return {'msg': getattr(self, 'msg', ''),
                'code': getattr(self, 'code', None),
                'data': getattr(self, 'data', {})}


class NotFoundError(BaseError):
    code = 404


class SecurityViolation(BaseError):
    pass


class AccessDenied(BaseError):
    msg = 'Access denied'


class ValidationError(BaseError):
    code = 400


class InvalidSessionError(BaseError):
    code = 401
    msg = 'Invalid session'


class ConflictError(BaseError):
    code = 409
    msg = 'Duplicate resource'



def extract_kw(request):
    return (request.args and dict((k, v) for (k, v) in request.args.items())) or \
            request.json or \
            (request.data and json.loads(request.data.decode('utf-8'))) or \
            request.form or \
            {}




from apphelpers.db.peewee import create_pgdb_pool, create_base_model, dbtransaction
from peewee import TextField

import settings

db = create_pgdb_pool(database=settings.DB_NAME, user=settings.DB_USER)
BaseModel = create_base_model(db)
dbtransaction = dbtransaction(db)


class Book(BaseModel):
    name = TextField()


def _add_book(name):
    Book.create(name=name)


def _add_book_loser(name):
    _add_book(name)
    loser  # will raise


def setup_module():
    Book.create_table()


def teardown_module():
    Book.drop_table()


def test_add_with_tr():
    add_book = dbtransaction(_add_book)
    name = "The Pillars of the Earth"
    add_book(name)
    names = [b.name for b in Book.select()]
    assert name in names

    add_book_loser = dbtransaction(_add_book_loser)
    name = "The Cathedral and the Bazaar"
    try:
        add_book_loser(name)
    except NameError:
        pass
    names = [b.name for b in Book.select()]
    assert name not in names

    add_book = dbtransaction(_add_book)
    name = "The Ego Trick"
    add_book(name)
    names = [b.name for b in Book.select()]
    assert name in names



import datetime
import decimal
import json
from functools import update_wrapper

import arrow
from flask import make_response, request, current_app, Response
from flask.json import JSONEncoder


def crossdomain(origin=None, methods=None, headers=None,
                max_age=10368000, attach_to_all=True,
                automatic_options=True):
    # http://flask.pocoo.org/snippets/56/
    if methods is not None:
        methods = ', '.join(sorted(x.upper() for x in methods))
    if headers is not None and not isinstance(headers, str):
        headers = ', '.join(x.upper() for x in headers)
    if not isinstance(origin, str):
        origin = ', '.join(origin)
    if isinstance(max_age, datetime.timedelta):
        max_age = max_age.total_seconds()

    def get_methods():
        if methods is not None:
            return methods

        options_resp = current_app.make_default_options_response()
        return options_resp.headers['allow']

    def decorator(f):
        def wrapped_function(*args, **kwargs):
            if automatic_options and request.method == 'OPTIONS':
                resp = current_app.make_default_options_response()
            else:
                resp = make_response(f(*args, **kwargs))
            if not attach_to_all and request.method != 'OPTIONS':
                return resp

            h = resp.headers

            h['Access-Control-Allow-Origin'] = origin
            h['Access-Control-Allow-Methods'] = get_methods()
            h['Access-Control-Max-Age'] = str(max_age)
            if headers is None:
                h['Access-Control-Allow-Headers'] = request.headers.get('Access-Control-Request-Headers', '')
            else:
                h['Access-Control-Allow-Headers'] = headers
            return resp

        f.provide_automatic_options = False
        return update_wrapper(wrapped_function, f)
    return decorator


def jsonify_unsafe(o):
    return json.dumps(o, cls=CustomJSONEncoder)


class CustomJSONEncoder(JSONEncoder):

    tz = None

    def default(self, obj):
        try:
            if isinstance(obj, (datetime.date, datetime.datetime)):
                if self.tz:
                    dt = arrow.get(obj)
                    return dt.to(self.tz).isoformat()
                return obj.isoformat()
            elif isinstance(obj, decimal.Decimal):
                return float(obj)
            iterable = iter(obj)
        except TypeError:
            pass
        else:
            return list(iterable)
        return JSONEncoder.default(self, obj)


def support_datetime_serialization(app, tz=None):
    CustomJSONEncoder.tz = tz
    app.json_encoder = CustomJSONEncoder
    return app


def add_cors_headers(resp):
    resp.headers['Access-Control-Allow-Origin'] = '*'
    resp.headers['Access-Control-Max-Age'] = '10368000'
    resp.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, OPTIONS, PATCH, DELETE'
    resp.headers['Access-Control-Allow-Headers'] = request.headers.get('Access-Control-Request-Headers', '')



import os
import requests
import time

import apphelpers.sessions as sessionslib

from converge import settings

base_url = 'http://127.0.0.1:8000/'
echo_url = base_url + 'echo'
secure_echo_url = base_url + 'secure-echo'
echo_groups_url = base_url + 'echo-groups'
pid_path = 'tests/run/app.pid'

sessiondb_conn = dict(host=settings.SESSIONSDB_HOST,
                      port=settings.SESSIONSDB_PORT,
                      password=settings.SESSIONSDB_PASSWD,
                      db=settings.SESSIONSDB_NO)
sessionsdb = sessionslib.SessionDBHandler(sessiondb_conn)


def gunicorn_setup_module():  # not working
    if os.path.exists(pid_path):
        os.remove(pid_path)
    cmd = f'gunicorn tests.service:__hug_wsgi__  -p {pid_path} -D'
    os.system(cmd)
    for i in range(10):
        if os.path.exists(pid_path):
            time.sleep(2)
            break


def gunicorn_teardown_module():
    if os.path.exists(pid_path):
        cmd = f'kill -9 `cat {pid_path}`'
        os.system(cmd)


def test_get():
    word = 'hello'
    url = echo_url + '/' + word
    assert requests.get(url).json() == word


def test_get_params():
    word = 'hello'
    url = echo_url + '/' + word
    params = {'word': word}
    assert requests.get(url, params=params).json() == word


def test_get_multi_params():
    nums = [3, 5]
    url = base_url + 'add'
    params = {'nums': nums}
    assert requests.get(url, params=params).json() == sum(nums)


def test_post():
    word = 'hello'
    url = echo_url
    assert requests.post(url, json={'word': word}).json() == word


def test_secure_echo():
    word = 'hello'
    headers = {'NoAuthorization': 'Header'}
    url = secure_echo_url + '/' + word
    resp = requests.get(url, headers=headers)
    assert resp.status_code == 401


def test_user_id():
    uid = 101
    d = dict(uid=uid, groups=[])
    sid = sessionsdb.create(**d)

    headers = {'Authorization': sid}

    word = 'hello'
    url = echo_url + '/' + word
    assert requests.get(url, headers=headers).json() == ('%s:%s' % (uid, word))

    url = base_url + 'me/uid'

    data = {'uid': None}
    resp = requests.post(url, json=data, headers=headers)
    assert resp.json() == uid

    data = {'uid': 1}  # invalid claim
    resp = requests.post(url, json=data, headers=headers)
    assert resp.json() == uid

    data = {'uid': uid}
    resp = requests.post(url, json=data, headers=headers)
    assert resp.json() == uid


def test_group_access():
    # 1. No group
    uid = 111
    groups = []
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = echo_groups_url

    headers = {'Authorization': sid}
    assert requests.get(url, headers=headers).status_code == 403

    # 2. Forbidden group
    uid = 112
    groups = ['noaccess-group']
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = echo_groups_url

    headers = {'Authorization': sid}
    assert requests.get(url, headers=headers).status_code == 403

    # 3. Access group
    uid = 113
    groups = ['access-group']
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)

    headers = {'Authorization': sid}
    assert requests.get(url, headers=headers).status_code == 200
    assert requests.get(url, headers=headers).json() == groups

def test_not_found():
    url = base_url + 'snakes/viper'
    assert requests.get(url).status_code == 404



# -*- coding: utf-8 -*-
from collections import namedtuple

from nose.tools import raises

import apphelpers.sessions as sessionslib
import settings

from apphelpers.errors import InvalidSessionError


sessiondb_conn = dict(host=settings.SESSIONSDB_HOST,
                      port=settings.SESSIONSDB_PORT,
                      password=settings.SESSIONSDB_PASSWD,
                      db=settings.SESSIONSDB_NO)
sessionsdb = sessionslib.SessionDBHandler(sessiondb_conn)


class data:
    pass


class state:
    pass


Session = namedtuple('Session', ['uid', 'groups', 'k', 'v'])
data.session = Session(987651, ['admin', 'member'], 'foo', 'bar')
data.session_email = 'noone@example.com'

#uid, groups, k, v = 987651, ['admin', 'member'], 'foo', 'bar'


def test_create():
    d = dict(uid=data.session.uid, groups=data.session.groups, extras=dict(email=data.session_email))
    sid = sessionsdb.create(**d)
    assert len(sid) == 43
    sid_new = sessionsdb.create(data.session.uid, data.session.groups)
    assert sid == sid_new
    state.sid = sid


def test_uodate():
    sid = state.sid
    k, v = data.session.k, data.session.v
    sessionsdb.update(sid, {k: v})
    d = sessionsdb.get(sid)
    assert d[k] == v
    assert d['email'] == data.session_email
    sessionsdb.remove_from_session(sid, k)
    d = sessionsdb.get(sid)
    assert k not in d


@raises(InvalidSessionError)
def test_delete():
    sessionsdb.destroy(state.sid)
    assert sessionsdb.get(state.sid)


def test_session_lookups():
    uids = range(10000, 10010)
    groups = ['grp1', 'grp2']
    for uid in uids:
        sid = sessionsdb.create(uid, groups)
        assert sessionsdb.sid2uidgroups(sid) == (uid, groups)
        sessionsdb.destroy(sid)
        raises(InvalidSessionError)(sessionsdb.get)(sid)



import hug

from apphelpers.rest.hug import APIFactory
from tests.app.endpoints import setup_routes

import settings


def make_app():
    router = hug.route.API(__name__)

    api_factory = APIFactory(router)
    # api_factory.setup_db_transaction(app.models.db)
    sessiondb_conn = dict(host=settings.SESSIONSDB_HOST,
                          port=settings.SESSIONSDB_PORT,
                          password=settings.SESSIONSDB_PASSWD,
                          db=settings.SESSIONSDB_NO)
    api_factory.setup_session_db(sessiondb_conn)
    setup_routes(api_factory)


make_app()



import hug
import hug.directives

from apphelpers.rest.hug import user_id

def echo(word, user: hug.directives.user=None):
    return '%s:%s' % (user.id, word) if user else word


def secure_echo(word, user: hug.directives.user=None):
    return '%s:%s' % (user.id, word) if user else word
secure_echo.login_required = True


def echo_groups(user: hug.directives.user=None):
    return user.groups
echo_groups.groups_required = ['access-group']
echo_groups.groups_forbidden = ['forbidden-group']


def add(nums: hug.types.multiple):
    return sum(int(x) for x in nums)


def get_my_uid(uid: user_id):
    return uid
get_my_uid.login_required = True


def get_snake(name):
    return None
get_snake.not_found_on_none = True


def setup_routes(factory):

    factory.get('/echo/{word}')(echo)
    factory.post('/echo')(echo)

    factory.get('/add')(add)

    factory.get('/secure-echo/{word}')(secure_echo)
    factory.get('/echo-groups')(echo_groups)

    factory.post('/me/uid')(get_my_uid)

    factory.get('/snakes/{name}')(get_snake)

    # ar_handlers = (None, arlib.create, None, arlib.get, arlib.update, None)
    # factory.map_resource('/resttest/', handlers=ar_handlers)



import datetime
import logging
import os

from functools import wraps
from enum import Enum

from peewee import DateTimeField, Model, Field
from playhouse.pool import PooledPostgresqlExtDatabase
from playhouse.shortcuts import model_to_dict


def set_peewee_debug():
    logger = logging.getLogger('peewee')
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)


class URLField(Field):
    field_type = 'text'

    def python_value(self, value):
        value = value.lower()
        if not value.startswith('http') or not '://' in value:
            raise TypeError("Not a http URL: %s" % value)
        return value


def create_pgdb_pool(
    host=None, database=None, user=None, password=None, max_connections=32
):
    return PooledPostgresqlExtDatabase(
        database=database,
        host=host,
        user=user,
        password=password,
        max_connections=max_connections,
        autorollback=True,
        register_hstore=False,
        stale_timeout=60 * 2,
    )  # 2 minutes


def create_base_model(db):
    class BaseModel(Model):
        class Meta:
            database = db
            only_save_dirty = True

        def to_dict(self, only=None, exclude=None, recurse=False):
            return model_to_dict(self, only=only, exclude=exclude, recurse=recurse)

    return BaseModel


created = lambda: DateTimeField(default=datetime.datetime.utcnow)


def dbtransaction(db):
    """
    wrapper that make db transactions automic
    note db connections are used only when it is needed (hence there is no usual connection open/close)
    """
    def wrapper(f):
        @wraps(f)
        def f_wrapped(*args, **kw):
            if not db.in_transaction():
                with db.connection_context():
                    with db.atomic():
                        result = f(*args, **kw)
            else:
                with db.atomic():
                    result = f(*args, **kw)
            return result
        return f_wrapped
    return wrapper


def enumify(TheModel, name_field="name", val_field="id"):
    """
    Converts a model rows into an enum
    Can be effective cache for mostly unchanging data.
    Limitation: No auto updates. If you update the model and you are using process manager like gunicorn you
    would need to restart to rnsure enums are updated

    eg.
    >>> class Week(BaseModel):
            day = CharField()
            num = IntField()

    >>> weekenum = enumify(Week, 'day', 'num')
    >>> print(weekenum.monday.num)
    """
    fields = getattr(TheModel, name_field), getattr(TheModel, val_field)
    data = list(
        (name.replace(" ", "_").lower(), v)
        for (name, v) in TheModel.select(*fields).tuples()
    )
    return Enum(TheModel.__name__, data)


def dbc(db):
    pid = os.getpid()
    return "[%s]: %s:%s" % (pid, len(db._in_use), db.max_connections)



import secrets
import _pickle as pickle
import redis
import hug

import apphelpers.context as context

from apphelpers.errors import InvalidSessionError


_SEP = ':'
session_key = ('session' + _SEP).__add__

rev_lookup_prefix = 'uid' + _SEP
rev_lookup_key = lambda uid: rev_lookup_prefix + str(uid)


class SessionDBHandler:

    def __init__(self, rconn_params):
        """
        rconn_params: redis connection parameters
        """
        self.rconn = redis.Redis(**rconn_params)

    def create(self, uid='', groups=None, extras=None, ttl=(30 * 24 * 60 * 60)):
        """
        groups: list
        extras (dict): each key-value pair of extras get stored into hset
        """
        if uid:
            sid = self.uid2sid(uid)
            if sid:
                return sid

        sid = secrets.token_urlsafe()
        key = session_key(sid)

        session_dict = {'uid': uid, 'groups': groups or []}
        if extras:
            session_dict.update(extras)
        session = {k: pickle.dumps(v) for k, v in session_dict.items()}
        self.rconn.hmset(key, session)

        if uid:
            rev_key = rev_lookup_key(uid)
            self.rconn.setex(rev_key, value=sid, time=ttl)
        self.rconn.expire(key, ttl)
        return sid

    def exists(self, sid):
        return self.rconn.exists(session_key(sid))

    def get(self, sid, keys=[]):
        s_values = self.rconn.hgetall(session_key(sid))
        if not s_values:
            raise InvalidSessionError()
        session = {k.decode(): pickle.loads(v) for k, v in s_values.items()}
        if keys:
            session = {k: session.get(k, None) for k in keys}
        return session

    def get_attribute(self, sid, attribute):
        value = self.rconn.hget(session_key(sid), attribute)
        return pickle.loads(value) if value else None

    def uid2sid(self, uid):
        sid = self.rconn.get(rev_lookup_key(uid))
        return sid.decode() if sid else None

    def get_for(self, uid):
        sid = self.uid2sid(uid)
        return self.get(sid) if sid else None

    def sid2uidgroups(self, sid):
        """
        => uid (int), groups (list)
        """
        session = self.get(sid, ['uid', 'groups'])
        return session['uid'], session['groups']

    def update(self, sid, keyvalues):
        sk = session_key(sid)
        keyvalues = {k: pickle.dumps(v) for k, v in list(keyvalues.items())}
        self.rconn.hmset(sk, keyvalues)

    def update_for(self, uid, keyvalues):
        sid = self.uid2sid(uid)
        return self.update(sid, keyvalues) if sid else None

    def update_attribute(self, sid, attribute, value):
        key = session_key(sid)
        self.rconn.hset(key, attribute, pickle.dumps(value))
        return True

    def remove_from_session(self, sid, keys):
        sk = session_key(sid)
        self.rconn.hdel(sk, keys)
        return True

    def destroy(self, sid):
        uid = self.sid2uidgroups(sid)[0]
        sk = session_key(sid)
        self.rconn.delete(sk)
        self.rconn.delete(rev_lookup_key(uid))
        return True

    def destroy_for(self, uid):
        sid = self.uid2sid(uid)
        return self.destroy(sid) if sid else None

    def destroy_all(self):
        keys = self.rconn.keys(session_key('*'))
        self.rconn.delete(*keys)
        keys = self.rconn.keys(rev_lookup_prefix + '*')
        self.rconn.delete(*keys)


def whoami(user: hug.directives.user):
    return user.to_dict()
whoami.login_required = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
    history = history_file.read()

requirements = [ ]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Scroll Tech",
    author_email='tech@scroll.in',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Common helper libraries for Python Apps",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='apphelpers',
    name='apphelpers',
    packages=find_packages(exclude=['tests']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/scrolltech/apphelpers',
    version='0.10.1',
    zip_safe=False,
)



import datetime
import json
import logging
import random
import sys

import settings

import urllib.parse as urllib
from functools import wraps, lru_cache

from flask import request, jsonify, make_response, Response
from flask import current_app
from .flaskutils import add_cors_headers, jsonify_unsafe
from apphelpers.errors import BaseError, AccessDenied, NotFoundError
from apphelpers.errors import InvalidSessionError
import apphelpers.users.sessions as sessionlib
import apphelpers.context as context

if settings.DB_TRANSACTIONS_ENABLED:
    from apphelpers.db.peewee import dbtransaction
else:
    dbtransaction = lambda f: f

SESSION_COOKIE_NAME = getattr(settings, 'SESSION_COOKIE_NAME', '__s')

cache = lru_cache()
cache_ttl = datetime.timedelta(0, (10*60))


def extract_kw(request):
    return (request.args and dict((k, v) for (k, v) in request.args.items())) or \
            request.json or \
            (request.data and json.loads(request.data.decode('utf-8'))) or \
            request.form or \
            {}


def flaskapi(app, f, jsonify_result=True):
    @wraps(f)
    def wrapper(*args, **kw):
        logger = current_app.logger
        context.current.uid = 0
        context.current.groups = []
        session_id = request.cookies.get(SESSION_COOKIE_NAME)
        if session_id:
            if request.environ['REQUEST_METHOD']:
                session_id = urllib.unquote(session_id)
            context.current.sid = session_id
        status_code = 200
        if request.method == 'OPTIONS':
            resp = app.make_default_options_response()
        else:
            kw.update(extract_kw(request))
            try:
                result = f(*args, **kw)
            except AccessDenied as err:
                result = err.to_dict()
                status_code = 403
                logger.exception('Access Denied error: ')
            except NotFoundError as err:
                result = err.to_dict()
                status_code = err.code or 404
                logger.exception('Object not found error: ')
            except BaseError as err:
                logger.exception('API Execution error: ')
                result = err.to_dict()
                status_code = getattr(err, 'code', 500)
            except Exception as err:
                err_id = str(random.random())[2:]
                logger.exception('Unhandled API Execution error [%s]: ', err_id)
                result = {'msg': ('Server error: ' + err_id)}
                status_code = getattr(err, 'code', 500)
                logger.error('[%s] parameters: %s', err_id, str(kw)[:400])
            if isinstance(result, dict):
                resp = jsonify(result)
            elif isinstance(result, Response):
                resp = result
                status_code = resp.status_code
            else:
                result = jsonify_unsafe(result) if jsonify_result else result
                resp = make_response(result)
            resp.status_code = status_code
        #add_cors_headers(resp)
        return resp
    return wrapper


def protected(f):
    @wraps(f)
    def wrapper(*args, **kw):
        session_id = kw.pop('_session_id', None) or hasattr(context.current, 'sid') and context.current.sid
        login_required = getattr(f, 'login_required', None)
        roles_required = getattr(f, 'roles_required', None)

        if login_required or roles_required:
            if session_id:
                uid, groups = sessionlib.sid2uidgroups(session_id)
            else:
                raise AccessDenied(msg='session not found')
        elif session_id:
            try:
                uid, groups = sessionlib.sid2uidgroups(session_id)
            except InvalidSessionError:
                uid, groups = None, []
        else:
            uid, groups = None, []
        context.set_context(sid=session_id, uid=uid, groups=groups)

        if roles_required and not set(context.current.groups).intersection(roles_required):
            raise AccessDenied(data=dict(groups=groups, roles_required=roles_required))

        return f(*args, **kw)
    return wrapper


def cached(f):
    if hasattr(f, 'cache'):
        cf = cache(f)
        cf.began = datetime.datetime.now()
        @wraps(f)
        def wrapper(*args, **kw):
            now = datetime.datetime.now()
            if (now - cf.began) > cache_ttl:
                cf.cache_clear()
                cf.began = now
            return cf(*args, **kw)
        wrapper.cache_info = cf.cache_info
        return wrapper
    return f


def api_factory(handler):
    return protected(cached(dbtransaction(handler)))


def add_url_rule(app, url, handler, methods, jsonify_result=True):
    # add debugging, inspection here
    print('%s -> %s %s' % (url, handler, str(methods)))
    if not 'OPTIONS' in methods:
        methods.append('OPTIONS')
    endpoint = url + '-' + str(methods)
    f = flaskapi(app, api_factory(handler), jsonify_result=jsonify_result)
    app.add_url_rule(url, endpoint, f, methods=methods)


def get_or_not_found(f):
    @wraps(f)
    def wrapper(*args, **kw):
        ret = f(*args, **kw)
        if ret is None:
            raise NotFoundError()
        return ret
    return wrapper


class RESTPublisher(object):
    """
    Expose a generic Python module as a RESTful service.
    This uses Flask, a micro-framework to expose required methods/functions.
    Current implementation does not implement partial edit or support for
    HTTP PATCH
    """
    def __init__(self, flask_app, api_urls_prefix='/api/'):
        """
        Initialized with an instance of flask app, uuid_type and api url prefix
        uuid_type denotes the type used by the resource as uuid
        e.g. int or str
        """
        self.app = flask_app
        self.urls_prefix = api_urls_prefix

    def map_resource(self, url, handlers, resource_id=('string', 'id')):
        """
        Maps a resource and its methods to URLs.
        All handlers are not required and may be passed as None
        handlers can be
            - object with http verbs as method names
                class Todos:
                    def get():
                    def post():
                class Todo:
                    def get():
                    def put():
                    def delete():
                todo_collection = Todos()
                todo_resource = Todo()
            - verb to function map
                {get: get_todo, post: create_todo}
            - list/tuple of functions handling http methods
                (get_todo, create_todo, ..)
        """
        id_type, id_name = resource_id
        collection_url = (self.urls_prefix + url) if not url.startswith('/') else url
        resource_url = collection_url + '<' + id_type + ':' + id_name + '>'

        if isinstance(handlers, (list, tuple)):
            get_collection, add_resource, replace_resource, get_resource, edit_resource, delete_resource = handlers
        else:
            raise NotImplemented

        if get_collection:
            add_url_rule(self.app, collection_url, get_collection, methods=['GET'])
        if add_resource:
            add_url_rule(self.app, collection_url, add_resource, methods=['POST'])
        if replace_resource:
            add_url_rule(self.app, collection_url, replace_resource, methods=['PUT'])
        if get_resource:
            get_resource_wrapped = get_or_not_found(get_resource)
            add_url_rule(self.app, resource_url, get_resource_wrapped, methods=['GET'])
        if edit_resource:
            add_url_rule(self.app, resource_url, edit_resource, methods=['PATCH'])
        if delete_resource:
            add_url_rule(self.app, resource_url, delete_resource, methods=['DELETE'])


class HTTPPublisher(object):
    """
    Expose functions/callables over HTTP.
    """
    def __init__(self, flask_app, api_urls_prefix='/api/'):
        self.app = flask_app
        self.urls_prefix = api_urls_prefix

    def add_mapping(self, url, handler, methods=['GET'], jsonify_result=True):
        """
        Add a mapping for a callable.
        """
        url = (self.urls_prefix + url) if not url.startswith('/') else url
        add_url_rule(self.app, url, handler, methods=methods, jsonify_result=jsonify_result)
        print('%s -> %s' % (url, handler))



from dataclasses import dataclass, asdict
from functools import wraps
from falcon import HTTPUnauthorized, HTTPForbidden, HTTPNotFound

import hug

from apphelpers.db.peewee import dbtransaction
from apphelpers.errors import InvalidSessionError
from apphelpers.sessions import SessionDBHandler


def phony(f):
    return f


def raise_not_found_on_none(f):
    if getattr(f, 'not_found_on_none', None) == True:
        @wraps(f)
        def wrapper(*a, **k):
            ret = f(*a, **k)
            if ret is None:
                raise HTTPNotFound('four o four')
            return ret
        return wrapper
    return f


@hug.directive()
def user_id(default=None, request=None, **kwargs):
    return request.context['user'].id


@hug.directive()
def user_name(default=None, request=None, **kwargs):
    return request.context['user'].name


@dataclass
class User:
    sid: str=None
    id: int=None
    name: str=None
    groups: tuple=()

    def to_dict(self):
        return asdict(self)

    def __bool__(self):
        return bool(self.id)


def setup_strict_context_setter(sessions):

    def set_context(token):

        uid, groups, name = None, [], ''

        if token:
            try:
                session = sessions.get(token, ['uid', 'name', 'groups'])
                uid, name, groups = session['uid'], session['name'], session['groups']
            except InvalidSessionError:
                raise HTTPUnauthorized('Invalid or expired session')

        return User(sid=token, id=uid, name=name, groups=groups)

    return set_context


def setup_context_setter(sessions):

    def set_context(response, request, context, module):
        """
        Only sets context based on session.
        Does not raise any error
        """
        uid, groups, name = None, [], ''
        token = request.get_header('Authorization')

        if token:
            try:
                session = sessions.get(token, ['uid', 'name', 'groups'])
                uid, name, groups = session['uid'], session['name'], session['groups']
            except InvalidSessionError:
                pass

        request.context['user'] = User(sid=token, id=uid, name=name, groups=groups)

    return set_context


class APIFactory:

    def __init__(self, router):
        self._router = router
        self.db_tr_wrapper = phony
        self.access_wrapper = phony
        self.secure_router = None

    def setup_db_transaction(self, db):
        self.db_tr_wrapper = dbtransaction(db)

    def setup_session_db(self, sessiondb_conn):
        """
        redis_conn_params: dict() with below keys
                           (host, port, password, db)
        """
        self.sessions = SessionDBHandler(sessiondb_conn)
        set_context = setup_context_setter(self.sessions)
        self.router = self._router.http(requires=set_context)
        set_context = setup_strict_context_setter(self.sessions)
        self.secure_router = self.router.http(requires=hug.authentication.token(set_context))

        def access_wrapper(f):
            """
            This is the authentication + authorization part
            """
            login_required = getattr(f, 'login_required', None)
            groups_required = getattr(f, 'groups_required', None)
            groups_forbidden = getattr(f, 'groups_forbidden', None)

            if login_required or groups_required or groups_forbidden:

                @wraps(f)
                def wrapper(request, *args, **kw):

                    user = request.context['user']

                    # this is authentication part
                    if not user.id:
                        raise HTTPUnauthorized('Invalid or expired session')

                    # this is authorization part
                    groups = set(user.groups)

                    if groups_required and not groups.intersection(groups_required):
                        raise HTTPForbidden('Unauthorized access')

                    if groups_forbidden and groups.intersection(groups_forbidden):
                        raise HTTPForbidden('Unauthorized access')

                    return f(*args, **kw)
            else:

                wrapper = f

            return wrapper

        self.access_wrapper = access_wrapper

    def choose_router(self, f):
        login_required = hasattr(f, 'login_required') and f.login_required
        return self.secure_router if login_required else self.router

    def build(self, method, method_args, method_kw, f):
        print(f'{method_args[0]} [{method.__name__.upper()}] => {f.__module__}:{f.__name__}')
        m = method(*method_args, **method_kw)
        f = raise_not_found_on_none(self.access_wrapper(self.db_tr_wrapper(f)))
        return m(f)

    def get(self, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            return self.build(router.get, a, k, f)
        return _wrapper

    def post(self, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            return self.build(router.post, a, k, f)
        return _wrapper

    def patch(self, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            return self.build(router.patch, a, k, f)
        return _wrapper

    def delete(self, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            return self.build(router.delete, a, k, f)
        return _wrapper

    def map_resource(self, url, resource=None, handlers=None, id_field='id'):
        if resource:
            raise NotImplementedError("Resource not supported yet")

        collection_url = (self.urls_prefix + url) if not url.startswith('/') else url
        resource_url = collection_url + '{' + id_field + '}'

        assert isinstance(handlers, (list, tuple)), "handlers should be list or tuple"
        get_collection, add_resource, replace_resource, get_resource, update_resource, delete_resource = handlers

        if get_collection:
            self.get(collection_url)(get_collection)
        if add_resource:
            self.post(collection_url)(add_resource)
        if replace_resource:
            self.put(resource_url)(replace_resource)
        if get_resource:
            # get_resource_wrapped = get_or_not_found(get_resource)
            self.get(resource_url)(get_resource)
        if update_resource:
            self.patch(resource_url)(update_resource)
        if delete_resource:
            self.delete(resource_url)(delete_resource)
